import os
import sys
import json
import multiprocessing
from contextlib import contextmanager



from ray.tune.suggest.variant_generator import generate_variants,flatten_resolved_vars

try:
    from hpogrid.components.defaults import *  
except:
    raise ImportError('ERROR: Cannot import hpogrid module. Try source setupenv.sh first.')



def get_base_path():
	if kHPOGridEnvPath not in os.environ:
		raise KeyError('{} environment variable not set.'
			'Try source setupenv.sh first.'.format(kHPOGridEnvPath))
	return os.environ[kHPOGridEnvPath]

def set_script_paths(proj_name, undo=False):

	local_job_scripts_path = os.path.join(get_base_path(), 'project', proj_name, 'scripts')
	grid_job_scripts_path = os.path.join(get_base_path(), 'project', 'scripts')

	script_paths = [local_job_scripts_path, grid_job_scripts_path]

	for path in script_paths:
		if (path in sys.path) and (undo==True):
			sys.path.remove(path)
			os.environ["PYTHONPATH"].replace(path+":","")
		if (path not in sys.path) and (undo==False):
			sys.path.append(path)
			os.environ["PYTHONPATH"] = path + ":" + os.environ.get("PYTHONPATH", "")


def get_project_path(proj_name):
	if kHPOGridProjPath in os.environ:
		proj_path = os.environ[kHPOGridProjPath]
	else:
		base_path = get_base_path()
		proj_base_path = os.path.join(base_path, 'project')
		proj_path = os.path.join(proj_base_path, proj_name)
	if not os.path.exists(proj_path):
		raise FileNotFoundError('Project "{}" not found.'.format(proj_name))
	return proj_path

def get_config(proj_name, config_type):
	project_path = get_project_path(proj_name)
	config_name = kConfig2FileMap[config_type]
	confg_path = os.path.join(project_path, 'config', config_name)
	if not os.path.exists(confg_path):
		raise FileNotFoundError('Missing {} config file: {}'.format(
			config_type.replace('_',''), confg_path))
	with open(confg_path) as config_file:
		config = json.load(config_file)
	return config

def get_project_config(proj_name):
	return get_config(proj_name, config_type='project')

def get_hpo_config(proj_name):
	return get_config(proj_name, config_type='hpo')

def get_grid_config(proj_name):
	return get_config(proj_name, config_type='grid')	

def get_model_config(proj_name):
	return get_config(proj_name, config_type='model')	

def get_search_space_config(proj_name):
	return get_config(proj_name, config_type='search_space')	

@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

def generate_hparams(search_space, n_point=1):
	hparams = []
	for i in range(num_samples):
		for resolved_vars, _ in generate_variants(search_space):
			hparams.append(flatten_resolved_vars(resolved_vars))
	return hparams


def get_physical_devices(device_type='GPU'):
	import tensorflow as tf
	physical_devices = tf.config.list_physical_devices('GPU')
	return physical_devices

def get_n_gpu():
	physical_devices = get_physical_devices('GPU')
	return len(physical_devices)

def get_n_cpu():
	return multiprocessing.cpu_count()


import os
import sys
import json
import argparse
from pdb import set_trace

import pandas as pd
from tabulate import tabulate

try:
    from hpogrid.components.PandaTaskManager import PandaTaskManager
    from hpogrid.components.defaults import *
except:
    raise ImportError('Cannot import hpogrid module. Try source setupenv.sh first.')

_h_test = ''

kDefaultExtraColumns = []
kSupportedExtraColumns = ['site', 'task_time_s', 'time_s', 'taskid']


class HPOReporter():

    def __init__(self, summary=None, metric=None, mode=None):
        self.summary = summary
        self.metric = metric
        self.mode = mode

    def is_valid_data(self, data):
        if not isinstance(data, dict):
            return False
        return all(key in kHPOGridMetadataFormat for key in data)

    def get_summary(self, data, extra=[]):
        summary = []
        if not self.is_valid_data(data):
            raise RuntimeError('Invalid format for hpo result summary.')
        hparams = data['hyperparameters']
        metric = data['metric']  
        mode =  data['mode']  
        # gather results from different trials
        for index in data['result']:
            trial_result = {}
            for hp in hparams:
                trial_result[hp] = data['result'][index][hp]
            trial_result[metric] = data['result'][index][metric]
            if 'time_s' in extra:
                trial_result['time_s'] = data['result'][index]['time_total_s']
            summary.append(trial_result)
        return summary, metric, mode

    def display(self):
        if self.summary is None:
            print('WARNING: No summary to display. Action aborted.')
            return None
        df = self.to_dataframe()
        print(tabulate(df, showindex=True, headers=df.columns, tablefmt="psql",stralign='center'))

    def from_data(self, data, extra=[]):
        self.summary, self.metric, self.mode = self.get_summary(data, extra)

    def from_summary(self, summary):
        self.summary = summary

    def from_json(self, file):
        with open(file) as json_file:
            self.summary = json.load(json_file)

    def to_dataframe(self):
        if self.summary is None:
            print('WARNING: No summary found. Cannot convert summary to dataframe.')
            return None
        df = pd.DataFrame(self.summary)

        if (self.metric != None) and (self.mode != None):
            if self.mode == 'min':
                sort_ascending = True
            elif self.mode == 'max':
                sort_ascending = False
            else:
                raise RuntimeError('Invalid mode {} for the metric {}'.format(self.mode, self.metric))
            df = df.sort_values(by=[self.metric], ascending=sort_ascending).reset_index(drop=True)
        return df

    def to_json(self, file='hpo_summary.json'):
        if self.summary is None:
            print('WARNING: No summary found. Cannot convert summary to json.')
            return None
        df = self.to_dataframe()
        with open(file, 'w') as output:
            json.dump(df.to_dict(), output)

    def to_html(self, file='hpo_result.html'):
        if self.summary is None:
            print('WARNING: No summary found. Cannot convert summary to html.')
            return None
        df = self.to_dataframe()
        html_text = df.to_html()
        with open(file, 'w') as outfile:
            outfile.write(html_text)
        return html_text

    def to_parallel_coordinate_plot(self, file='hpo_result_parallel_coordinate_plot.html'):
        if self.summary is None:
            print('WARNING: No summary found. Cannot convert summary to parallel coordinate plot.')
            return None

        import hiplot as hip
        html_text = hip.Experiment.from_iterable(self.summary).to_html()
        if file is not None:
            with open(file, 'w') as outfile:
                outfile.write(html_text)
        return html_text

    def to_csv(self, file='hpo_result.csv'):
        if self.summary is None:
            print('WARNING: No summary found. Cannot convert summary to csv.')
            return None
        df = self.to_dataframe()
        csv = df.to_csv(file, encoding='utf-8')
        return csv


class HPOTaskHandle(HPOReporter):

    def __init__(self):
        super().__init__()
        self.taskmgr = PandaTaskManager()
        if len(sys.argv) > 1:
            self.run_parser()

    def get_parser(self, method=None):
        parser = argparse.ArgumentParser(description='Tool for managing HPO result',
                    formatter_class=argparse.RawDescriptionHelpFormatter)    
        parser.add_argument('proj_name', help='Name of project')
        parser.add_argument('-l','--limit', help=_h_test , type=int,
                            default=1000)
        parser.add_argument('-d', '--days', help=_h_test, type=int,
                            default=30)
        parser.add_argument('-n', '--taskname', help=_h_test)
        parser.add_argument('-r', '--range', help=_h_test)
        parser.add_argument('-e', '--extra', nargs='+',
            default=kDefaultExtraColumns)
        parser.add_argument('-j', '--to_json', action='store_true', help=_h_test)
        parser.add_argument('-t', '--to_html', action='store_true', help=_h_test)
        parser.add_argument('-c', '--to_csv', action='store_true', help=_h_test)
        parser.add_argument('-p', '--to_pcp', action='store_true', help=_h_test)
        parser.add_argument('-o', '--outname', help=_h_test, default='hpo_result')
        return parser

    def run_parser(self):
        parser = self.get_parser()
        args = parser.parse_args(sys.argv[2:])
        self.show(args)

    def show(self, args):
        params = vars(args)
        params['status'] = 'done'
        params['metadata'] = True
        proj_name = params.pop('proj_name', '')
        outname = params.pop('outname', None)
        datasets = self.taskmgr.query_tasks(**params)
        # filter tasks by jetitaskid range
        if args.range is not None:
            datasets = self.taskmgr.filter_range(datasets, args.range)
        col = ['computingsite', 'jobs_metadata', 'jeditaskid']
        datasets = self.taskmgr.filter_datasets(datasets, col)

        if args.extra:
            for extra in args.extra:
                if not extra in kSupportedExtraColumns:
                    raise ValueError('Unsupported extra column {}'.format(extra))

        self.summary, self.metric, self.mode = self.format_summary(datasets, proj_name, args.extra)
        self.display()

        if args.to_json:
            self.to_json(outname+'.json')
        if args.to_html:
            self.to_html(outname+'.html')
        if args.to_csv:
            self.to_csv(outname+'.csv')    
        if args.to_pcp:
            self.to_parallel_coordinate_plot(outname+'_parallel_coordinate_plot.html')   

    def filter_invalid_datasets(self, datasets, proj_name):
        filtered_datasets = []
        for dataset in datasets:
            metadata = self.extract_metadata(dataset)
            if not metadata:
                continue
            if not self.is_valid_data(metadata):
                continue
            if metadata['title'] != proj_name:
                continue
            filtered_datasets.append(dataset)
        if len(filtered_datasets) == 0:
            print('INFO: No results found.')
            sys.exit(1)
        metrics = set([metadata['metric'] for data in filtered_datasets])
        if not len(metrics)==1 :
            raise RuntimeError('HPO results from different tasks consist of different metrics: {}'.format(metrics))
        hparams = [metadata['hyperparameters'] for data in filtered_datasets]
        if not all(hp == hparams[0] for hp in hparams):
            raise RuntimeError('HPO results from different tasks consist of different hyperparameter space')
        return filtered_datasets

    def extract_metadata(self, dataset):
        if 'jobs_metadata' in dataset:
            keys = list(dataset['jobs_metadata'].keys())
            if keys:
                return dataset['jobs_metadata'][keys[0]]
            else:
                return None
        else:
            return None

    def format_summary(self, datasets, proj_name, extra=kDefaultExtraColumns):
        summary = []
        valid_datasets = self.filter_invalid_datasets(datasets, proj_name)

        metric =  None
        mode = None

        for data in valid_datasets:
            metadata = self.extract_metadata(data)
            task_summary, metric, mode = self.get_summary(metadata, extra)
            for ts in task_summary:
                if 'taskid' in extra:
                    ts['taskid'] = data['jeditaskid']
                if 'task_time_s' in extra:
                    ts['task_time_s'] = metadata['task_time_s']
                if 'site' in extra:
                    ts['site'] = data['computingsite']
            summary += task_summary
        return summary, metric, mode



"""
	Module for formatting tables, texts and figures used in hpogrid
"""
from tabulate import tabulate
import pandas as pd

version = '0.1.2'

kDefaultTableStyle = 'psql'
kDefaultStrAlign = 'left'

class ColorCode():
	RED = '\033[0;91m'
	GREEN = '\033[92m'
	YELLOW = '\033[93m'
	BLUE = '\033[94m'
	MAGENTA = '\033[95m'
	CYAN = '\033[96m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'
	BLINK = '\033[5m'

def create_table(data, columns=None, indexed=True, transpose=False,
	tableformat=kDefaultTableStyle, stralign=kDefaultStrAlign):
	df = pd.DataFrame(data, columns=columns)
	if transpose:
		df = df.transpose()
	table = tabulate(df, showindex=indexed, headers=df.columns, 
	    tablefmt=tableformat,
	    stralign=stralign)
	return table



kHPOGridEnvPath = 'HPOGRID_BASE_PATH'
kHPOGridProjPath = 'HPOGRID_PROJECT_PATH'

kProjectConfigName = 'project_config.json'
kModelConfigName = 'model_config.json'
kSearchSpaceConfigName = 'search_space.json'
kHPOConfigName = 'hpo_config.json'
kGridConfigName	 = 'grid_config.json'

kConfig2FileMap = {
	'project': kProjectConfigName,
	'hpo': kHPOConfigName,
	'search_space': kSearchSpaceConfigName,
	'model': kModelConfigName,
	'grid': kGridConfigName
}


kSearchAlgorithms = ['hyperopt', 'skopt', 'bohb', 'ax', 'tune', 'random', 'grid', 'bayesian']
kSchedulers = ['asynchyperband', 'bohbhyperband', 'pbt']
kMetricMode = ['min', 'max']


kDefaultSearchAlgorithm = 'random'
kDefaultMetric = 'accuracy'
kDefaultMode = 'max'
kDefaultScheduler = 'asynchyperband'
kDefaultLogDir = './log'
kDefaultTrials = 100
kDefaultStopping = '{"training_iteration": 1}'
kDefaultSchedulerParam = '{}'
kDefaultAlgorithmParam = '{"max_concurrent": 4}'
kDefaultModelParam = '{}'

kGPUGridSiteList = ['ANALY_MANC_GPU_TEST', 'ANALY_QMUL_GPU_TEST', 'ANALY_MWT2_GPU',
'ANALY_BNL_GPU_ARC', 'ANALY_INFN-T1_GPU']


kDefaultGridSite = 'ANALY_MANC_GPU_TEST'
kDefaultContainer = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'
kDefaultContainer2 = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/clcheng/hyperparameter-optimization-on-the-grid:latest'
kDefaultContainer3 = 'docker://gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'

kDefaultOutDS = 'user.${{RUCIO_ACCOUNT}}.hpogrid.{HPO_PROJECT_NAME}.out.$(date +%Y%m%d%H%M%S)'

kGridSiteMetadataFileName = 'userJobMetadata.json'

kHPOGridMetadataFormat = ['title', 'metric', 'mode', 'task_time_s', 'result', 'hyperparameters', 'best_config']



import os
import json
import shutil
import subprocess

import tarfile
from pdb import set_trace

#import re
import fnmatch

try:
    from hpogrid.utils import stylus
except:
    raise ImportError('Cannot import hpogrid module. Try source setupenv first.')


def extract_inputDS(in_path='/ctrdata/', out_path='/ctrdata/'):
    tarfiles = [ f for f in os.listdir(in_path) if f.endswith('tar.gz')]
    for f in tarfiles:
        tar = tarfile.open(f, "r:gz")
        print('untaring the file {}'.format(f))
        tar.extractall(path=out_path)
        inputDS = tar.getnames()
        tar.close()


    return inputDS

def remove_inputDS(inputDS):
    for ds in inputDS:
        if os.path.isfile(ds):
            os.remove(ds)
        elif os.path.isdir(ds):
            shutil.rmtree(ds)

def show_grid_site_info(name_filter=None, gpu=True, active=True, site_type=['analysis', 'unified'],
    info = ['state', 'status', 'maxinputsize', 'maxmemory', 'maxtime']):
   grid_site_info = extract_grid_site_info(name_filter, gpu, active, site_type, info)
   print(stylus.create_table(grid_site_info, transpose=True))


def get_grid_site_list(name_filter=None, gpu=True, active=True, site_type=['analysis', 'unified']):
    grid_site_info = extract_grid_site_info(name_filter, gpu, active, site_type)
    return grid_site_info.keys()

def extract_grid_site_info(name_filter=None, gpu=True, active=True, site_type=['analysis', 'unified'],
    info = ['state', 'status', 'maxinputsize', 'maxmemory', 'maxtime']):
    '''
    retrieve some basic information of PanDA grid sites
    '''
    try:
      jsonfileLocation = os.environ['ALRB_cvmfs_repo'] + '/sw/local/etc/agis_schedconf.json'
    except:
      jsonfileLocation = '/cvmfs/atlas.cern.ch/repo/sw/local/etc/agis_schedconf.json'

    with open(jsonfileLocation,'r') as jsonfile:
      jsondata = json.load(jsonfile)

    if name_filter is None:
        name_filter = '*'

    grid_site_info = {}

    for site in jsondata:
        # filter site names (could also match jsondata[site]['panda_resource'] instead)
        if not fnmatch.fnmatch(site, name_filter):
            continue
        # filter non-active grid sites
        if active and (not jsondata[site]['state'] == 'ACTIVE'):
            continue
        # no good indicator of a GPU site yet will just judge on site name
        if gpu and (not 'GPU' in jsondata[site]['panda_resource']):
            continue
        if jsondata[site]['type'] not in site_type:
            continue
        grid_site_info[site] = {key: jsondata[site][key] for key in info}

    return grid_site_info